import os
import pandas as pd
from tqdm import tqdm
from .Comparative_condition import *
from .Single_Condition import *
from .utils import *


def analise_LR(lrpaths, 
                genes=None, 
                tf_genes=None, 
                out_path=None, 
                sep=',', 
                threshold=0, 
                colors=None, 
                out_file=None, 
                output_fmt="html_document", 
                sel_columns=['source','target','gene_A','gene_B','type_gene_A','type_gene_B','MeanLR'], 
                org='hsa', comparison=None, filtered_net=False):
    
    """
    Core engine to generate report. Here we perform all the computation related to pyCrossTalkeR

    Parameters
    ----------
    lrpaths :
        Paths of single condition LR data
    genes :
        list of genes to be considered in the sankey plots
    out_path :
        output directory path
    sep :
        character used on csv
    threshold :
        percentage of edges to be pruned
    colors :
        celltypes colorscheme
    out_file :
        output file names
    output_fmt :
        rmarkdown render output format parameter
    sel_columns :
        columns from data
    
    Returns
    -------
    Rmarkdown report all objects from each step
    
    """
    
    data = read_lr_single_condition(lrpaths, 
                                    sel_columns, 
                                    out_path, 
                                    sep, 
                                    colors)

    print("Create a Differential Table")
    if len(lrpaths) > 1:
        data = create_diff_table(data, out_path, comparison)
        data = fisher_test_cci(data, 'LRScore', out_path, comparison)
        data = mannwitu_test_cci(data, 'LRScore', out_path, comparison)
        data = filtered_graphs(data, out_path)

    print("Calculating CCI Ranking")
    data = ranking(data, out_path, sel_columns=sel_columns, slot="graphs")
    print("Calculating GCI Ranking")
    data = ranking(data, out_path, sel_columns=sel_columns, slot="graphs_ggi")
    print("Network Analysis Done")

    with open(os.path.join(out_path, "LR_data.pkl"), "wb") as f:
        pickle.dump(data, f)

    return(data)

# paths = {
#     'CTR': "Test_Data/CTR_LR.csv",
#     'EXP': "Test_Data/EXP_LR.csv"
# }
# output = "output/"
# data = analise_LR(paths, out_path=output, org="hsa")




import pandas as pd
import networkx as nx

def create_diff_table(data, out_path, comparison=None):
    """
    Read the lrobject and generate the comparative tables

    Parameters
    ----------
    data :
        LRObj with single condition
    out_path :
        output path
    
    Returns
    -------
    LRObject
    
    """

    def process_pair(exp_table, ctr_table):
        tmp_data = pd.merge(exp_table, ctr_table, on='allpair', how='outer')
        tmp_data[['ligpair', 'recpair']] = tmp_data['allpair'].str.split('@', expand=True)
        tmp_data['LRScore_x'] = tmp_data['LRScore_x'].fillna(0)
        tmp_data['LRScore_y'] = tmp_data['LRScore_y'].fillna(0)
        tmp_data['LRScore'] = tmp_data['LRScore_x'] - tmp_data['LRScore_y']
        
        final_data = tmp_data[tmp_data['LRScore'] != 0].copy()
        final_data['type_gene_A'] = final_data['type_gene_A_x'].combine_first(final_data['type_gene_A_y'])
        final_data['type_gene_B'] = final_data['type_gene_B_x'].combine_first(final_data['type_gene_B_y'])
        final_data['gene_A'] = final_data['gene_A_x'].combine_first(final_data['gene_A_y'])
        final_data['gene_B'] = final_data['gene_B_x'].combine_first(final_data['gene_B_y'])
        final_data['source'] = final_data['source_x'].combine_first(final_data['source_y'])
        final_data['target'] = final_data['target_x'].combine_first(final_data['target_y'])
        final_data['cellpair'] = final_data['source'] + "@" + final_data['target']
        final_data['interaction_type'] = final_data['type_gene_A'] + final_data['type_gene_B']
        
        final_data['interaction_type'] = final_data['interaction_type'].replace({
            'LigandReceptor': 'LR',
            'ReceptorTranscription Factor': 'RTF',
            'Transcription FactorLigand': 'TFL'
        })
        
        final = (final_data[~final_data['type_gene_A'].str.contains('Transcription') & 
                            ~final_data['type_gene_B'].str.contains('Transcription')]
                 .groupby('cellpair')['LRScore'].sum().reset_index())
        
        final[['u', 'v']] = final['cellpair'].str.split('@', expand=True)
        final.dropna(subset=['u', 'v'], inplace=True)
        
        raw_inter = final_data.loc[~final_data['type_gene_A'].str.contains('Transcription') & 
                                   ~final_data['type_gene_B'].str.contains('Transcription'),
                                   'cellpair'].value_counts()
        
        raw_inter = raw_inter.reindex(final['cellpair']).fillna(0)
        freq = (raw_inter - raw_inter.min()) / (raw_inter.max() - raw_inter.min()) + 0.1
        final['freq'] = freq.values
        final['pair'] = final['cellpair']
        
        return final_data, final

    if comparison is not None:
        for pair in comparison:
            ctr_name, exp_name = pair[1], pair[0]
            cmp_name = f"{exp_name}_x_{ctr_name}"
            exp_table, ctr_table = data['tables'][exp_name], data['tables'][ctr_name]
            final_data, final = process_pair(exp_table, ctr_table)
            data['tables'][cmp_name] = final_data
            
            G = nx.DiGraph()
            for _, row in final.iterrows():
                G.add_edge(row['u'], row['v'], LRScore=row['LRScore'], freq=row['freq'], weight=row['LRScore'], inter=row['freq'])
            
            data['graphs'][cmp_name] = G
            
            G_ggi = nx.DiGraph()
            for _, row in final_data.iterrows():
                G_ggi.add_edge(row['ligpair'], row['recpair'], LRScore=row['LRScore'], weight=row['LRScore'], inter=row['LRScore'])
            
            data['graphs_ggi'][cmp_name] = G_ggi
    else:
        ctr_name = list(data['tables'].keys())[0]
        ctr_table = data['tables'][ctr_name]
        for exp_name in list(data['tables'].keys())[1:]:
            cmp_name = f"{exp_name}_x_{ctr_name}"
            exp_table = data['tables'][exp_name]
            final_data, final = process_pair(exp_table, ctr_table)
            data['tables'][cmp_name] = final_data
            
            G = nx.DiGraph()
            for _, row in final.iterrows():
                G.add_edge(row['u'], row['v'], LRScore=row['LRScore'], freq=row['freq'], weight=row['LRScore'], inter=row['freq'])
            
            data['graphs'][cmp_name] = G
            
            G_ggi = nx.DiGraph()
            for _, row in final_data.iterrows():
                G_ggi.add_edge(row['ligpair'], row['recpair'], LRScore=row['LRScore'], weight=row['LRScore'], inter=row['LRScore'])
            
            data['graphs_ggi'][cmp_name] = G_ggi
    
    return data



#'
#'Run and Generate all LR Downstream analysis
#'
#'This function loads the single conditions LR outputs and return the LR network
#`based analysis.
#'It assumes that the table present the following columns Ligand,
#` Ligand.Cluster,Receptor,Receptor.Cluster and MeanLR/another
#'measure
#'@slot graphs All Cell Cell Interaction Networks
#'@slot tables All tables from single condition
#'@slot max_iter  Max meanLR from all
#'@slot max_nodes All Celltype in the experiment
#'@slot coords  Cell Cell Interaction Plots
#'@slot colors  Cell type colors
#'@slot rankings Ranking of cells and Genes
#'@slot loadings  CCI values to remove multiple times genes
#'@slot pca  PCA results
#'@slot annot  Annotation Results


import attr

@attr.s
class LRObj:
    """
    This function loads the single conditions LR outputs and return the LR network It assumes that the table present the following columns Ligand, measure

    Attributes
    ----------
    graphs :
        All Cell Cell Interaction Networks
    tables :
        All tables from single condition
    max_iter :
        Max meanLR from all
    max_nodes :
        All Celltype in the experiment
    coords :
        Cell Cell Interaction Plots
    colors :
        Cell type colors
    rankings :
        Ranking of cells and Genes
    loadings :
        CCI values to remove multiple times genes
    pca :
        PCA results
    annot :
        Annotation Results
    
    """

    graphs = attr.ib(default=attr.Factory(dict))
    graphs_ggi = attr.ib(default=attr.Factory(dict))
    tables = attr.ib(default=attr.Factory(dict))
    max_iter = attr.ib(default=0)
    max_nodes = attr.ib(default=0)
    coords = attr.ib(default=None)
    colors = attr.ib(default=None)
    rankings = attr.ib(default=attr.Factory(list))
    loadings = attr.ib(default=attr.Factory(list))
    pca = attr.ib(default=attr.Factory(list))
    annot = attr.ib(default=attr.Factory(list))
    stats = attr.ib(default=attr.Factory(list))



import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import igraph
import itertools
from scipy.stats import fisher_exact, MonteCarloMethod, mannwhitneyu
import pickle
from itertools import combinations
import os
import networkx as nx
import re

#'Ranking the most interactive gene (ligand or receptor)
#'
#'@param data lrobject
#'@param out_path to save the lrobject with ranking
#'@param sel_columns columns to consider
#'@param slot slot of the networks graphs_ggi to gene cell interaction and abs
#'@import tibble
#'@import utils
#'@import dplyr
#'@return list
#'@importFrom tidyr %>%
#'@importFrom stats prcomp
#'@NoRd


def ranking(data, out_path, sel_columns, slot="graphs"):
    """
    Ranking the most interactive gene (ligand or receptor)

    Parameters
    ----------
    data :
        lrobject
    out_path :
        to save the lrobject with ranking
    sel_columns :
        columns to consider
    slot :
        slot of the networks graphs_ggi to gene cell interaction and abs
    
    Returns
    -------
    list
    
    """

    sc = StandardScaler()
    slot_data = data.get(slot, {})
    rankings = data.get('rankings', {})
    pca_data = data.get('pca', {})
    stats_data = data.get('stats', {})
    model = PCA(n_components= 2)
    for graph_name, graph in slot_data.items():
        if "_x_" in graph_name:  # Signed Analysis
            components = list(nx.connected_components(graph.to_undirected()))
            all_both = None
            
            for comp in components:
                subgraph = graph.subgraph(comp)
                if all_both is None:
                    all_both = ranking_net(subgraph, mode=False)
                else:
                    tmp = ranking_net(subgraph, mode=False)
                    all_both = pd.concat([all_both, tmp], ignore_index=True)
            
            if "_ggi" in slot:
                all_both = comparative_pagerank(rankings, slot, graph_name, all_both)
                all_both = comparative_med(rankings, slot, graph_name, all_both)
                rankings[f"{graph_name}_ggi"] = all_both
                pca_data_frame = all_both.drop(columns=['nodes']).loc[:, (all_both != 0).any(axis=0)]
                pca_data_frame = sc.fit_transform(pca_data_frame)
                pca = model.fit_transform(pca_data_frame)
                pca_data[f"{graph_name}_ggi"] = pca
            else:
                all_both = comparative_pagerank(rankings, slot, graph_name, all_both)
                all_both = comparative_med(rankings, slot, graph_name, all_both)
                rankings[graph_name] = all_both
                pca_data_frame = all_both.drop(columns=['nodes']).loc[:, (all_both != 0).any(axis=0)]
                pca_data_frame = sc.fit_transform(pca_data_frame)
                pca = model.fit_transform(pca_data_frame)
                pca_data[graph_name] = pca
        else:  # Unsigned Analysis
            components = list(nx.connected_components(graph.to_undirected()))
            all_data = None
            
            for comp in components:
                subgraph = graph.subgraph(comp)
                if all_data is None:
                    all_data = ranking_net(subgraph)
                else:
                    tmp = ranking_net(subgraph)
                    all_data = pd.concat([all_data, tmp], ignore_index=True)
            
            if "_ggi" in slot:
                final_data = None
                table = data['tables'][graph_name]
                cls = pd.unique(pd.concat([table[sel_columns[0]], table[sel_columns[1]]]))
                for i in cls:
                    all_eq = pd.unique(pd.concat([
                        table['ligpair'][table[sel_columns[0]] == i],
                        table['recpair'][table[sel_columns[1]] == i]
                    ]))
                    edges = pd.DataFrame(list(itertools.combinations(all_eq, 2)), columns=['u', 'v'])
                    edges['LRScore'] = 0.0
                    if final_data is None:
                        final_data = edges
                    else:
                        final_data = pd.concat([final_data, edges], ignore_index=True)
                
                tmp_tbl = table[['ligpair', 'recpair', 'LRScore']]
                all1 = pd.concat([tmp_tbl, final_data])
                tmp_net = nx.from_pandas_edgelist(all1, source='ligpair', target='recpair', edge_attr='LRScore', create_using=nx.DiGraph())
                pg = nx.pagerank(tmp_net, weight='LRScore')
                all_data['Pagerank'] = all_data['nodes'].map(pg)
                rankings[f"{graph_name}_ggi"] = all_data
                pca_data_frame = all_data.drop(columns=['nodes']).loc[:, (all_data != 0).any(axis=0)]
                pca = model.fit_transform(pca_data_frame)
                pca_data[f"{graph_name}_ggi"] = pca 
            else:
                pg = nx.pagerank(graph, weight='LRScore')
                all_data['Pagerank'] = all_data['nodes'].map(pg)
                rankings[graph_name] = all_data
                pca_data_frame = all_data.drop(columns=['nodes']).loc[:, (all_data != 0).any(axis=0)]
                pca_data_frame = sc.fit_transform(pca_data_frame)
                pca = model.fit_transform(pca_data_frame)
                pca_data[graph_name] = pca

    data['rankings'] = rankings
    data['pca'] = pca_data
    data['stats'] = stats_data

    with open(os.path.join(out_path, "LR_data_final.pkl"), "wb") as f:
        pickle.dump(data, f)
    
    return data



#'Ranking the most interactive gene (ligand or receptor)
#'
#'@param rankings tables lrobject
#'@param slotname slot of the networks graphs_ggi to gene cell interaction and abs
#'@param graphname graph comparison name
#'@param curr.rkg ranking table
#'@importFrom tidyr %>%
#'@import stringr
#'@return list
#'@NoRd
#' Network Ranking method
#'
#'@param graph lrobject
#'@param mode  is TRUE if is comparive mode
#'@return list
#'@import igraph
#'@importFrom tidyr %>%
#'@NoRd

def ranking_net(graph, mode=True):
    """
    Network Ranking method

    Parameters
    ----------
    graph :
        lrobject
    mode :
        is TRUE if is comparive mode
    
    Returns
    -------
    list
    
    """

    nodes = list(graph.nodes)
    
    if not mode:
        # Positive weights subgraph
        pos_edges = [(u, v) for u, v, d in graph.edges(data=True) if d['weight'] > 0]
        pos_subgraph = graph.edge_subgraph(pos_edges)
        deg_in_pos = dict(pos_subgraph.in_degree()) if isinstance(graph, nx.DiGraph) else dict(pos_subgraph.degree())
        deg_out_pos = dict(pos_subgraph.out_degree()) if isinstance(graph, nx.DiGraph) else dict(pos_subgraph.degree())
        
        # Negative weights subgraph
        neg_edges = [(u, v) for u, v, d in graph.edges(data=True) if d['weight'] < 0]
        neg_subgraph = graph.edge_subgraph(neg_edges)
        deg_in_neg = dict(neg_subgraph.in_degree()) if isinstance(graph, nx.DiGraph) else dict(neg_subgraph.degree())
        deg_out_neg = dict(neg_subgraph.out_degree()) if isinstance(graph, nx.DiGraph) else dict(neg_subgraph.degree())
        
        # Ensure nodes exist in the dictionaries
        deg_in_pos = {node: deg_in_pos.get(node, 0) + 1 for node in nodes}
        deg_out_pos = {node: deg_out_pos.get(node, 0) + 1 for node in nodes}
        deg_in_neg = {node: deg_in_neg.get(node, 0) + 1 for node in nodes}
        deg_out_neg = {node: deg_out_neg.get(node, 0) + 1 for node in nodes}
        
        centrality_table = pd.DataFrame({
            'nodes': nodes,
            'Listener': [round(deg_in_pos[node] - deg_in_neg[node], 2) for node in nodes],
            'Influencer': [round(deg_out_pos[node] - deg_out_neg[node], 2) for node in nodes]
        })
    else:
        # Calculating degrees
        deg_in_pos = dict(graph.in_degree()) if isinstance(graph, nx.DiGraph) else dict(graph.degree())
        deg_out_pos = dict(graph.out_degree()) if isinstance(graph, nx.DiGraph) else dict(graph.degree())
        
        # Calculating betweenness centrality with absolute weights
        abs_weights_graph = graph.copy()
        for u, v, d in abs_weights_graph.edges(data=True):
            d['weight'] = abs(d['weight'])
        bet = nx.betweenness_centrality(abs_weights_graph, weight='weight', normalized= False)
        
        centrality_table = pd.DataFrame({
            'nodes': nodes,
            'Listener': [round(deg_in_pos[node], 2) for node in nodes],
            'Influencer': [round(deg_out_pos[node], 2) for node in nodes],
            'Mediator': [round(bet[node], 2) for node in nodes]
        })
    
    centrality_table.fillna(0, inplace=True)
    return centrality_table


#'Ranking the most interactive gene (ligand or receptor)
#'
#'@param rankings tables lrobject
#'@param slotname slot of the networks graphs_ggi to gene cell interaction and abs
#'@param graphname graph comparison name
#'@param curr.rkg ranking table
#'@importFrom tidyr %>%
#'@import stringr
#'@return list
#'@NoRd

def comparative_pagerank(rankings, slotname, graphname, curr_rkg):
    """
    Ranking the most interactive gene (ligand or receptor)

    Parameters
    ----------
    rankings :
        tables lrobject
    slotname :
        slot of the networks graphs_ggi to gene cell interaction and abs
    graphname :
        graph comparison name
    curr.rkg :
        ranking table
    
    Returns
    -------
    list
    
    """
    
    p_f1 = p_f2 = 0.5  # probability to be at disease
    allnodes = pd.DataFrame(curr_rkg['nodes'], columns=['nodes'])
    if '_filtered' in graphname:
        curr = re.sub('_filtered', '', graphname)
        curr = re.split('_x_', curr)
        p_ctr = curr[1]
        q_exp = curr[0]
    else:
        curr = re.split('_x_', graphname)
        p_ctr = curr[1]
        q_exp = curr[0]

 
    if "_ggi" in slotname:
        p = rankings[p_ctr + '_ggi'][['nodes','Pagerank']]
        q = rankings[q_exp + '_ggi'][['nodes','Pagerank']]

        

    else:
        p = rankings[p_ctr][['nodes', 'Pagerank']].loc[rankings[q_exp]['Pagerank'].index]
        q = rankings[q_exp][['nodes', 'Pagerank']]

    
    p.columns = ['nodes', 'p_ctr']
    q.columns = ['nodes', 'p_dis']

    final = pd.merge(allnodes,p, on='nodes', how='left')
    final = pd.merge(final, q, on='nodes', how='left')

    final['p_ctr'] = final['p_ctr'].fillna(0)
    final['p_dis'] = final['p_dis'].fillna(0)
    
    alpha = 0.01
    final['p_ctr'] = final['p_ctr'] + alpha
    final['p_dis'] = final['p_dis'] + alpha

    final['p_ctr'] = final['p_ctr'] / final['p_ctr'].sum()
    final['p_dis'] = final['p_dis'] / final['p_dis'].sum()

    p = final['p_ctr']
    q = final['p_dis']

    pc = p * p_f1 + q * p_f2
    pcontrol = (p_f1 * p) / pc
    pdisease = (p_f2 * q) / pc

    final_result = np.log(pdisease / pcontrol)  
    
    curr_rkg['Pagerank'] = final_result
    return curr_rkg


#'Delta betweenness the most interactive gene (ligand or receptor)
#'
#'@param rankings tables lrobject
#'@param slotname slot of the networks graphs_ggi to gene cell interaction and abs
#'@param graphname graph comparison name
#'@param curr.rkg ranking table
#'@importFrom tidyr %>%
#'@import stringr
#'@return list
#'@NoRd

def comparative_med(rankings, slotname, graphname, curr_rkg):
    """
    Delta betweenness the most interactive gene (ligand or receptor)

    Parameters
    ----------
    rankings :
        tables lrobject
    slotname :
        slot of the networks graphs_ggi to gene cell interaction and abs
    graphname :
        graph comparison name
    curr.rkg :
        ranking table
    
    Returns
    -------
    list
    
    """
     
    allnodes = pd.DataFrame(curr_rkg['nodes'], columns=['nodes'])
    if '_filtered' in graphname:
        curr = re.sub('_filtered', '', graphname)
        curr = re.split('_x_', curr)
        p_ctr = curr[1]
        q_exp = curr[0]
    else:
        curr = re.split('_x_', graphname)
        p_ctr = curr[1]
        q_exp = curr[0]

   
    if "_ggi" in slotname:
        p = rankings[p_ctr + '_ggi'][['nodes', 'Mediator']]
        q = rankings[q_exp + '_ggi'][['nodes', 'Mediator']]
    else:
        p = rankings[p_ctr][['nodes', 'Mediator']].loc[rankings[q_exp][['nodes', 'Mediator']].index]
        q = rankings[q_exp][['nodes', 'Mediator']]

    p.columns = ['nodes', 'm_ctr']
    q.columns = ['nodes', 'm_exp']

    final = pd.merge(allnodes,p, on='nodes', how='left')
    final = pd.merge(final, q, on='nodes', how='left')

    final['m_ctr'] = final['m_ctr'].fillna(0)
    final['m_exp'] = final['m_exp'].fillna(0)

    curr_rkg['Mediator'] = final['m_exp'] - final['m_ctr']
    return curr_rkg


def add_node_type(df):
    """
    Adding genetype to the gene names to distinguish biological function

    Parameters
    ----------
    df :
        dataframe with interaction data
    
    Returns
    -------
    df
    
    """

    df['gene_A'] = df.apply(lambda row: f"{row['gene_A']}|L" if row['type_gene_A'] == "Ligand" else row['gene_A'], axis=1)
    df['gene_A'] = df.apply(lambda row: f"{row['gene_A']}|R" if row['type_gene_A'] == "Receptor" else row['gene_A'], axis=1)
    df['gene_A'] = df.apply(lambda row: f"{row['gene_A']}|TF" if row['type_gene_A'] == "Transcription Factor" else row['gene_A'], axis=1)
    
    df['gene_B'] = df.apply(lambda row: f"{row['gene_B']}|L" if row['type_gene_B'] == "Ligand" else row['gene_B'], axis=1)
    df['gene_B'] = df.apply(lambda row: f"{row['gene_B']}|R" if row['type_gene_B'] == "Receptor" else row['gene_B'], axis=1)
    df['gene_B'] = df.apply(lambda row: f"{row['gene_B']}|TF" if row['type_gene_B'] == "Transcription Factor" else row['gene_B'], axis=1)
    
    return df


def fisher_test_cci(data, measure, out_path, comparison=None):
    """
    Evaluate Differences in the edge proportion

    Parameters
    ----------
    data :
        data
    measure :
        intensity
    out_path :
        save path
    
    Returns
    -------
    data (lr_object) with fisher stats
    
    """

    if comparison is not None:
        for pair in comparison:
            ctr_name = pair[1]
            exp_name = pair[0]
            
            c = data['tables'][ctr_name].groupby('cellpair').size().reset_index(name='measure')
            e = data['tables'][exp_name].groupby('cellpair').size().reset_index(name='measure')

            joined = pd.merge(c, e, on='cellpair', how='outer', suffixes=('_ctr', '_exp'))
            
            pvals = []
            measure_ctr_sum = joined['measure_ctr'].sum()
            measure_exp_sum = joined['measure_exp'].sum()
            for idx, row in joined.iterrows():
                ctotal = joined['measure_ctr'].sum() - row['measure_ctr']
                etotal = joined['measure_exp'].sum() - row['measure_exp']
                matrix = np.array([[row['measure_exp'], etotal], [row['measure_ctr'], ctotal]])

                n_resamples = 1000
                np.random.seed(42)  # For reproducibility

                matrix_data = np.hstack([
                    np.repeat(0, matrix[0, 0]),
                    np.repeat(1, matrix[0, 1]),
                    np.repeat(2, matrix[1, 0]),
                    np.repeat(3, matrix[1, 1]),
                ])

                bootstrap_p_values = []
                for _ in range(n_resamples):
                    resampled_data = np.random.choice(matrix_data, size=matrix_data.size, replace=True)
                    resampled_table = np.array([
                        [np.sum(resampled_data == 0), np.sum(resampled_data == 1)],
                        [np.sum(resampled_data == 2), np.sum(resampled_data == 3)]
                    ])
                    _, p_value = fisher_exact(resampled_table)
                    bootstrap_p_values.append(p_value)

                # Estimate the empirical p-value
                bootstrap_p_values = np.array(bootstrap_p_values)
                empirical_p_value = np.mean(bootstrap_p_values <= p_value)

                # Fisher's exact test
                _, p_value_1 = fisher_exact(matrix)

                # print(p_value_1)
                # print(empirical_p_value)
                
                # Log odds ratio
                odds_ratio = row['measure_exp'] / row['measure_ctr'] if row['measure_ctr'] != 0 else np.nan
                lodds = np.log2(odds_ratio) if odds_ratio > 0 else np.nan
                
                pvals.append({
                    'cellpair': row['cellpair'],
                    'p_value': p_value_1,
                    'lodds': lodds
                })

            pval_df = pd.DataFrame(pvals)
            data['stats'][f'{exp_name}_x_{ctr_name}'] = pval_df

        with open(os.path.join(out_path, "LR_data_final.pkl"), "wb") as f:
            pickle.dump(data, f)

        return data

    else:
        if len(data['tables']) >= 2:
            c = data['tables']['CTR'].groupby('cellpair').size().reset_index(name='measure')

            for i in range(1, len(data['tables'])):
                if '_x_' not in list(data['tables'].keys())[i]:
                    e = data['tables'][list(data['tables'].keys())[i]].groupby('cellpair').size().reset_index(name='measure')

                    # Merge control and experimental data on 'cellpair'
                    joined = pd.merge(c, e, on='cellpair', how='inner', suffixes=('_ctr', '_exp'))
                    
                    pvals = []
                    measure_ctr_sum = joined['measure_ctr'].sum()
                    measure_exp_sum = joined['measure_exp'].sum()
                    for idx, row in joined.iterrows():
                        ctotal = measure_ctr_sum - row['measure_ctr']
                        etotal = measure_exp_sum - row['measure_exp']
                        matrix = np.array([[row['measure_exp'], etotal], [row['measure_ctr'], ctotal]])
                        odds_ratio, p_value = fisher_exact(matrix, alternative="two-sided")
    
                        # Monte Carlo resampling (if B > 0)
                        B = 1000
                        np.random.seed(42)  # For reproducibility
                        if B > 0:
                            count = 0
                            for _ in range(B):
                                shuffled = np.random.permutation(matrix.flatten()).reshape(matrix.shape)
                                if fisher_exact(shuffled, alternative="two-sided")[1] <= p_value:
                                    count += 1
                            perm_p_value = (count + 1) / (B + 1)  # Avoid zero probability
                        else:
                            perm_p_value = None

                        lodds = np.log2(odds_ratio) if odds_ratio > 0 else None  # Compute log odds ratio

                        pvals.append({
                            'cellpair': row['cellpair'],
                            'p_value': p_value,
                            'lodds': lodds
                        })

                    pval_df = pd.DataFrame(pvals)
                    data['stats'][f'{list(data["tables"].keys())[i]}_x_{list(data["tables"].keys())[0]}'] = pval_df

            with open(os.path.join(out_path, "LR_data_final.pkl"), "wb") as f:
                pickle.dump(data, f)

            return data


def filtered_graphs(data, out_path):
    """
    Filter comparison graphs by statistics

    Parameters
    ----------
    data :
        data
    out_path :
        save path
    
    Returns
    -------
    data (lr_object) with fisher comparison graphs
    
    """

    temp = {}
    for name, graph in data['graphs'].items():
        if '_x_' in name:
            h = [edge[0] for edge in graph.edges()]
            f = [edge[1] for edge in graph.edges()]

            stats = data['stats'][name]
            
            significant_edges = stats[stats['p_value'] <= 0.05]
            
            significant_edge_pairs = [
                (h[i], f[i]) for i, edge_pair in enumerate(zip(h, f))
                if f"{h[i]}@{f[i]}" in significant_edges['cellpair'].values
            ]

            filtered_graph = graph.edge_subgraph(significant_edge_pairs).copy()
            temp[name] = filtered_graph

    for name in temp:   
        data['graphs'][f"{name}_filtered"] = temp[name]

    with open(os.path.join(out_path, "LR_data_final.pkl"), "wb") as f:
        pickle.dump(data, f)

    return data


def mannwitu_test_cci(data, measure, out_path, comparison=None):
    """
    Evaluate Differences in the edge strength

    Parameters
    ----------
    data :
        data
    measure :
        intensity
    out_path :
        save path
    
    Returns
    -------
    data (lr_object) with mannwittu stats
    
    """

    lcellpair = {}
    for key, df in data['tables'].items():
        lcellpair[key] = df['cellpair'].unique()

    if comparison:
        for pair in comparison:
            ctr_name, exp_name = pair

            results = []
            for cellpair in np.unique(np.concatenate(list(lcellpair.values()))):
                c = data['tables'][ctr_name].loc[data['tables'][ctr_name]['cellpair'] == cellpair, ['allpair', measure]]
                e = data['tables'][exp_name].loc[data['tables'][exp_name]['cellpair'] == cellpair, ['allpair', measure]]

                merged = pd.merge(c, e, on='allpair', how='outer').fillna(0)
                
                stat, p_value = mannwhitneyu(merged[measure + '_x'], merged[measure + '_y'], alternative='two-sided')
                
                results.append({
                    'cellpair': cellpair,
                    'statistic': stat,
                    'p_value': p_value
                })

            data['stats'][f'{exp_name}_x_{ctr_name}:MannU'] = pd.DataFrame(results)
    
    else:
        c = data['tables']['CTR']
        
        for key, df in data['tables'].items():
            if key != 'CTR':

                results = []
                for cellpair in np.unique(np.concatenate(list(lcellpair.values()))):
                    c = data['tables']['CTR'].loc[data['tables']['CTR']['cellpair'] == cellpair, ['allpair', measure]]
                    e = df.loc[df['cellpair'] == cellpair, ['allpair', measure]]

                    merged = pd.merge(c, e, on='allpair', how='outer').fillna(0)

                    stat, p_value = mannwhitneyu(merged[measure + '_x'], merged[measure + '_y'], alternative='two-sided')

                    results.append({
                        'cellpair': cellpair,
                        'statistic': stat,
                        'p_value': p_value
                    })

                data['stats'][f'{key}_x_CTR:MannU'] = pd.DataFrame(results)

    return data



import pandas as pd
import numpy as np
import networkx as nx
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from .utils import *

def read_lr_single_condition(lrpaths, sel_columns, out_path="/tmp/", sep=",", colors=None):
    """
    This function loads the single conditions LR outputs and use it to generate the report data and it`s object It assumes that the table presents the following columns Ligand, Ligand.Cluster, Receptor,Receptor.Cluster and MeanLR/another measure

    Parameters
    ----------
    lrpaths :
        Named vector with the lrpaths of each output
    sel_columns :
        selected columns
    out_path :
        Path to deposit the results
    sep :
        character used to divide the columns on input file
    colors :
        colorlist
    
    Returns
    -------
    LRObject
    
    """

    data = {}
    graphs = {}
    graphs_ggi = {}
    unif_celltypes = []

    for cond, lrpath in lrpaths.items():
        # Reading data
        if isinstance(lrpath, str):
            data1 = pd.read_csv(lrpath, sep=sep)    
        elif isinstance(lrpath, pd.DataFrame):
            data1 = lrpath.copy()
        else:
            raise ValueError("lrpath must be either a file path or a DataFrame")
        
        if not (data1['gene_A'].str.contains(r'\|').sum() > 0):
            data1 = add_node_type(data1)
        
        # Uniting columns
        data1['cellpair'] = data1[sel_columns[0]] + '@' + data1[sel_columns[1]]
        if 'Ligand' in data1[sel_columns[4]].unique():
            data1['ligpair'] = data1[sel_columns[0]] + '/' + data1[sel_columns[2]]
            data1['recpair'] = data1[sel_columns[1]] + '/' + data1[sel_columns[3]]
            data1['allpair'] = data1['ligpair'] + '@' + data1['recpair']
            data1[['ligpair', 'recpair']] = data1['allpair'].str.split('@', expand=True)
        else:
            data1[sel_columns[4]], data1[sel_columns[5]] = data1[sel_columns[5]], data1[sel_columns[4]]
            data1[sel_columns[3]], data1[sel_columns[4]] = data1[sel_columns[4]], data1[sel_columns[3]]
            data1['ligpair'] = data1[sel_columns[0]] + '/' + data1[sel_columns[2]]
            data1['recpair'] = data1[sel_columns[1]] + '/' + data1[sel_columns[3]]
            data1['allpair'] = data1['ligpair'] + '@' + data1['recpair']
            data1[['ligpair', 'recpair']] = data1['allpair'].str.split('@', expand=True)

        unif_celltypes += data1[sel_columns[0]].unique().tolist()
        unif_celltypes += data1[sel_columns[1]].unique().tolist()

        data1['LRScore'] = data1[sel_columns[-1]]

        final = data1.groupby('cellpair')['LRScore'].sum().reset_index()

        aux = final['cellpair'].str.split('@', expand=True)
        final['pair'] = final['cellpair']
        final[['u', 'v']] = aux

        filtervar = data1[sel_columns[4]].str.contains('Transcription') | data1[sel_columns[5]].str.contains('Transcription')
        raw_inter = data1.loc[~filtervar, 'cellpair'].value_counts()
        freq = (raw_inter[final['pair']].values - min(raw_inter)) / (max(raw_inter) - min(raw_inter)) + 0.1
        final['freq'] = freq

        # Convert to NetworkX
        graph1 = nx.DiGraph()
        for index, row in final.iterrows():
            graph1.add_edge(row['u'], 
                            row['v'], 
                            LRScore=row['LRScore'], 
                            freq=row['freq'], 
                            weight=row['LRScore'], 
                            inter=row['freq'])  # Add thickness (inter)

        graph2 = nx.DiGraph()
        for index, row in data1.iterrows():
            graph2.add_edge(row['ligpair'], 
                            row['recpair'], 
                            LRScore=row['LRScore'], 
                            mean=row['LRScore'], 
                            weight=row['LRScore'], 
                            inter=row['LRScore'])  # Add thickness (inter)

        data[cond] = data1
        graphs[cond] = graph1
        graphs_ggi[cond] = graph2

    # Create a full graph
    template = nx.complete_graph(len(set(unif_celltypes)))
    c  = nx.circular_layout(template)

    coords = {node: c[i] for i, node in enumerate(sorted(set(unif_celltypes), key=lambda x: x.lower()), start=0)}

    if colors is None:
        matplot_colors = plt.cm.get_cmap('Paired', len(set(unif_celltypes)))
        colors = {node: mcolors.to_hex(matplot_colors(i)) for i, node in enumerate(sorted(set(unif_celltypes), key=lambda x: x.lower()), start=0)}

    lr = {"graphs": graphs,
          "graphs_ggi": graphs_ggi,
          "tables": data,
          "colors": colors, 
          "coords": coords,
          "rankings": {},
          "pca" : {},
          "stats" : {}} 

    return lr



import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as mcolors
import seaborn as sns
import plotly.colors as pc
import plotly.graph_objects as go
from plotnine import *
from adjustText import adjust_text
from gprofiler import GProfiler
from sankeyflow import Sankey


def plot_cci(graph, colors, plt_name, coords, pg, emax=None, leg=False, low=25, high=75, ignore_alpha=False, log=False, efactor=8, vfactor=12, vnames=True,figsize=None):
    """
    This function does a CCI plot

    Parameters
    ----------
    graph :
        Paths of single condition LR data
    colors :
        Cell type (Cluster) Colors
    plt_name :
        Plot Name (Title)
    coords :
        object coordinates
    emax :
        Max MeanLR across the all inputs, if its not defined, the method going to consider the max find within a sample
    leg :
        Set color legend
    low :
        Lower threshold: This parameter low and high defines the edges
    high :
        Higher threshould which will be filtered. Edges within the interval [low\,high] are filtered.
    ignore_alpha :
        not include transparency on the plot
    log :
        logscale the interactions
    efactor :
        edge scale factor
    vfactor :
        edge scale factor
    vnames :
        remove vertex labels
    pg :
        pagerank values

    Returns
    -------
    Python default plot
    
    """

    # Check Maximal Weight
    if emax is None:
        emax = 0
        for _, _, d in graph.edges(data=True):
            weight = d.get('weight', 0)
            if isinstance(weight, (int, float)):
                emax = max(emax, abs(weight))

    # Create colormap
    colors_list = sns.color_palette("coolwarm", 201)  # Adjust to match the R colormap
    col_pallet_colors = [colors_list[i] for i in range(201)]
    col_pallet_colors[10] = '#B8b9ba'  # Expand the palette range

    # Scale coordinates
    coords_array = np.array(list(coords.values()))

    if coords_array.shape[0] != 1:
        coords_mean = (np.mean(coords_array[:, 0]), np.mean(coords_array[:, 1]))
        coords_std = (np.std(coords_array[:, 0]), np.std(coords_array[:, 1]))
        coords_scale = {node: tuple((coord - mean) / std for coord, mean, std in zip(coords[node], coords_mean, coords_std)) for node in coords}
    else:
        coords_scale = coords

    coords_scale = {key : (coords_scale[key][0] * 1.5, coords_scale[key][1] * 1.5) for key in coords_scale}

    # Calculate edge colors and alpha
    edge_colors = []
    alpha = []

    for u, v, d in graph.edges(data=True):
        weight = d.get('weight', 0)
        we = np.round(np.interp(weight, [-emax, emax], [1, 200]))
        edge_colors.append(col_pallet_colors[int(we)])
        alpha_cond = low < d.get('inter', 0) < high and not np.isnan(d.get('inter', 0) < high)
        alpha.append(0 if alpha_cond else d.get('inter', 0) < high)

    # Set edge attributes
    for u, v, d in graph.edges(data=True):
        d['color'] = [(c[0], c[1], c[2], a) for c, a in zip(edge_colors, alpha)]
        if log:
            d['width'] = np.log2(1 + d.get('inter', 0)) * efactor if d.get('inter', 0) != 0 else 0
        else:
            d['width'] = d.get('inter', 0) * efactor if d.get('inter', 0) != 0 else 0
        d['arrow_size'] = 0.4
        d['arrow_width'] = d['width'] + 0.8
        d['loop_angle'] = np.nan

    node_colors = [str(colors.get(node)) for node in graph.nodes()]
    node_sizes = [size*500 for size in pg]

    # Plot the graph
    if figsize is None:
        figsize = (6,6)
    fig, ax = plt.subplots(figsize=figsize)
    
    nx.draw(graph, pos=coords_scale, edge_color=edge_colors, node_color=node_colors, node_size=node_sizes,
            width=[d['width'] for _, _, d in graph.edges(data=True)],
            arrows=True, arrowsize=30, arrowstyle='-|>',
            connectionstyle='arc3,rad=0.3', with_labels=vnames)
    ax.set_xlim(-4, 4)
    ax.set_ylim(-4, 4)

    # Node Pagerank legend
    if pg is not None:
        min_pg, max_pg = min(pg), max(pg)
        legend1 = ax.legend(loc='lower left', title="Pagerank",
                handles=[plt.Line2D([], [], linestyle='', marker='o', markersize=v / vfactor, markerfacecolor='black', markeredgecolor='none') for v in [min_pg, (min_pg + max_pg) / 2, max_pg]],
                labels=[round(min_pg, 2), round((min_pg + max_pg) / 2, 2), round(max_pg, 2)],  bbox_to_anchor=(0.8, 0))

    # Thickness legend
    non_zero_inter_edges = [d['inter'] for _, _, d in graph.edges(data=True) if d.get('inter', 0) != 0]
    if non_zero_inter_edges:
        e_wid_sp = [round(min(non_zero_inter_edges), 2), round(min(non_zero_inter_edges) + (emax / 2), 2), round(emax, 2)]
        legend2 = ax.legend(e_wid_sp, title='Percentage of \nthe interactions', title_fontsize='small', loc='upper left', bbox_to_anchor=(0.8, 0.4))

    ax.add_artist(legend1)
    ax.add_artist(legend2)
    
    ax.set_title(plt_name)
    # Show the plot
    plt.tight_layout()
    plt.show()


def plot_pca_LR_comparative(lrobj_tblPCA, pca_table, dims=(1, 2), ret=False, ggi=True, include_tf=False, gene_types="all"):
    """
    This function is a proxy to the PCA plot in comparative conditions

    Parameters
    ----------
    lrobj_tblPCA :
        LRobject table with all data
    pca_table :
        table entry
    dims :
        PCA dims
    ret :
        return plot
    ggi :
        GGI mode
    include_tf :
        intracellular option
    gene_types :
        filter option of genes
    
    Returns
    -------
    Python default plot
    
    """
    
    pca_plot = {}
    # Extract PCA results and create a DataFrame
    pca_result = lrobj_tblPCA['pca'][pca_table]
    pca_df = lrobj_tblPCA['rankings'][pca_table]
    pca_df[['PC1', 'PC2']] = pca_result # Adjust dims to be zero-indexed
    pca_df = pca_df.set_index("nodes")

    if ggi:
        # Filter for LR or TF
        if gene_types == "LR":
            result_split_names = [name for name in pca_df.index if "|R" in name or "|L" in name]
        elif gene_types == "TF":
            result_split_names = [name for name in pca_df.index if "|TF" in name]
        else:
            result_split_names = pca_df.index.tolist()

        pca_df = pca_df.loc[result_split_names]

        # Mapping Table
        if include_tf:
            map_df = pd.DataFrame(pca_df.index, columns=["gene"])
            map_df["mapping"] = map_df["gene"].apply(lambda gene: "Receptor" if "|R" in gene else ("Ligand" if "|L" in gene else "Transcription Factor"))
            color_groups = ["#f8756b", "#00b835", "#619cff"]
        else:
            l_mapping = lrobj_tblPCA["tables"][pca_table.replace('_ggi', '')][["ligpair", "type_gene_A"]].rename(columns={"ligpair": "gene", "type_gene_A": "mapping"}).drop_duplicates()
            r_mapping = lrobj_tblPCA["tables"][pca_table.replace('_ggi', '')][["recpair", "type_gene_B"]].rename(columns={"recpair": "gene", "type_gene_B": "mapping"}).drop_duplicates()
            map_df = pd.concat([l_mapping, r_mapping]).drop_duplicates().reset_index(drop=True)
            map_df = map_df[map_df["gene"].isin(pca_df.index)]
            color_groups = ["#f8756b", "#00b835"]

        # Merge mapping info with PCA data
        pca_df = pca_df.merge(map_df, left_index=True, right_on="gene")

        #Threshold

        sdev_x = pca_df['PC1'].std()
        sdev_y = pca_df['PC2'].std()
        ver_zx = np.abs(pca_df['PC1']) >= (4 * sdev_x)
        ver_zy = np.abs(pca_df['PC2']) >= (4 * sdev_y)

        # Plotting
        x_max = max(abs(pca_df['PC1']))
        y_max = max(abs(pca_df['PC2']))

        pca_df['PC1'] = -pca_df['PC1']
        plt.figure(figsize=(10, 7))
        sns.scatterplot(x='PC1', y='PC2', data=pca_df, s=20, hue='mapping', palette=color_groups)

        # Adjust text labels to avoid overlap
        texts = []
        for i, gene in enumerate(pca_df['gene']):
            if ver_zx.iloc[i] or ver_zy.iloc[i]: 
                texts.append(plt.text(pca_df.loc[pca_df['gene'] == gene, 'PC1'].values[0], pca_df.loc[pca_df['gene'] == gene, 'PC2'].values[0], gene, fontsize=8,  bbox=dict(facecolor='white', edgecolor='black', boxstyle='round,pad=0.3')))
        
        adjust_text(texts,
                    force_text=(1.0, 2.0))

        plt.xlim(-x_max, x_max)
        plt.ylim(-y_max, y_max)
        plt.xlabel(f'PC{dims[0]}')
        plt.ylabel(f'PC{dims[1]}')
        plt.title(pca_table, y=1.08)
        plt.legend(title='Gene Type')
        plt.grid(True)

        plt.axhline(0, linestyle='--', color='gray')
        plt.axvline(0, linestyle='--', color='gray')
        
        plt.show()

        pca_plot[pca_table] = plt
        
    else:
        # No GGI
        x_max = max(abs(pca_df['PC1']))
        y_max = max(abs(pca_df['PC2']))

        pca_df['PC1'] = -pca_df['PC1']
        plt.figure(figsize=(10, 7))
        sns.scatterplot(x='PC1', y='PC2', data=pca_df)

        # Adjust text labels to avoid overlap
        texts = []
        for i, gene in enumerate(pca_df.index):
            texts.append(plt.text(pca_df.loc[gene, 'PC1'], pca_df.loc[gene, 'PC2'], gene, fontsize=8))
        
        adjust_text(texts, arrowprops=dict(arrowstyle='->', color='red'))

        plt.xlim(-x_max, x_max)
        plt.ylim(-y_max, y_max)
        plt.xlabel(f'PC{dims[0]}')
        plt.ylabel(f'PC{dims[1]}')
        plt.title(pca_table)
        plt.grid(True)

        # Set x and y axis intervals
        plt.xticks(np.arange(-np.ceil(x_max), np.ceil(x_max) + 1))
        plt.yticks(np.arange(-np.ceil(y_max), np.ceil(y_max) + 1))

        plt.axhline(0, linestyle='--', color='gray')
        plt.axvline(0, linestyle='--', color='gray')
        plt.show()

        pca_plot[pca_table] = plt

    if ret:
        return pca_plot
    

def plot_bar_rankings(data, table_name, ranking, type = None, filter_sign = None, mode = "cci", top_num = 10):
    """
    This function generates the barplot for a given network ranking on the CGI level. Further, the genes can be filtered by selected gene types to filter the plot.

    Parameters
    ----------
    data_object :
        LRobject with all data

    table_name :
        name of the ranking table

    ranking :
        name of the network ranking to use

    type :
        gene type (L,R,TF, LR/RL, RTF/TFR, LTF/TFL)

    filter_sign :
        show all (NULL), only positive (pos), or only negativ (neg) results
    
    Returns
    -------
    Python default plot

    """

    if '_x_' in table_name:
        rankings_table = data['rankings'][table_name]

        if type != None:
            if len(type) == 1:
                rankings_table = rankings_table[rankings_table['nodes'].str.contains('\\|' + type)]
            elif len(type) == 2:
                print(type)
                if type == 'TF':
                    rankings_table = rankings_table[rankings_table['nodes'].str.contains('\\|' + type)]
                else:
                    rankings_table = rankings_table[rankings_table['nodes'].str.contains('\\|' + type + '|\\|' + type[::-1])]
            elif len(type) == 3:
                if type == 'RTF' or type == 'TFR':
                    rankings_table = rankings_table[rankings_table['nodes'].str.contains('\\|' + 'RTF' + '|\\|' + 'TFR')]
                elif type == 'LTF' or type == 'TFL':
                    rankings_table = rankings_table[rankings_table['nodes'].str.contains('\\|' + 'LTF' + '|\\|' + 'TFL')]

        rankings_table = rankings_table.sort_values(by=ranking)
        
        if mode == 'cgi':
            rankings_table = pd.concat([rankings_table.head(top_num), rankings_table.tail(top_num)])
            # rankings_table.loc[rankings_table[ranking].abs().nlargest(20).index]
        else:
            pass

        if filter_sign == 'pos':
            rankings_table = rankings_table[rankings_table['ranking'] > 0]
        elif filter_sign == 'neg':
            rankings_table = rankings_table[rankings_table['ranking'] < 0]


        if rankings_table.empty:
            return "No entries with provided Filters."

        rankings_table['signal'] = ['negative' if x < 0 else 'positive' for x in rankings_table[ranking]]

        custom_palette = {'positive': '#FF6E00', 'negative':'#00FFFF'}  # Orange and Blue

        # Plot
        plt.figure(figsize=(8, 6))
        sns.barplot(x=ranking, y='nodes', data=rankings_table, hue='signal', dodge=False, palette=custom_palette)
        plt.title(f"Ranking for {table_name}")
        plt.xlabel(ranking)
        plt.ylabel('Nodes')

        # Set x-axis tick intervals
        max_val = rankings_table[ranking].max()
        min_val = rankings_table[ranking].min()
        ticks = np.linspace(min_val, max_val, num=5)  # Adjust 'num' for more/less intervals
        plt.xticks(ticks, [f'{tick:.2f}' for tick in ticks])

        # Invert y-axis to have highest values at the top
        plt.gca().invert_yaxis()

        # Show the legend only once
        handles, labels = plt.gca().get_legend_handles_labels()
        plt.legend(handles, labels, loc='lower right')

        plt.grid(True, linestyle='--', linewidth=0.5)
        plt.gca().set_axisbelow(True)
        # Adjust layout and show plot
        plt.tight_layout()
        plt.show()


def plot_sankey(lrobj_tbl, target = None, ligand_cluster = None, receptor_cluster = None, plt_name = None, threshold = 50, tfflag = True):
    """
    This function selected genes sankey plot

     Parameters
    ----------
    lrobj_tbl :
        LRobject table with all data

    target :
        gene

    ligand_cluster :
        Ligand Clusters

    receptor_cluster :
        Receptor Clusters

    plt_name :
        plot title

    threshold :
        top_n n value
    
    Returns
    -------
    Python default plot

    """

    lrobj_tbl = lrobj_tbl[(lrobj_tbl['type_gene_A'] == "Ligand") & (lrobj_tbl['type_gene_B'] == "Receptor")]

    if target is not None:
        if len(target.split('|')) > 1:
            target_type = str(target.split('|')[1])
            if target_type == 'R':
                if lrobj_tbl['gene_B'].str.contains('\\|').any():
                    pass
                else:
                    target = target.split('|')[0]
                data = lrobj_tbl[lrobj_tbl['gene_B'] == target]
            elif target_type == 'L':
                if lrobj_tbl['gene_A'].str.contains('\\|').any():
                    pass
                else:
                    target = target.split('|')[0]
                data = lrobj_tbl[lrobj_tbl['gene_A'] == target]
        else:
            data = lrobj_tbl[lrobj_tbl['allpair'].str.contains(target)]
    else:
        data = lrobj_tbl

    
    if ligand_cluster is not None:
        data = data[data['source'].isin(ligand_cluster)]
    
    if receptor_cluster is not None:
        data = data[data['target'].isin(receptor_cluster)]

    color_palette = ['#00BFC4', '#FF3E3E']

    
    if len(data) >= 1:
        cat_cols = ['source', 'gene_A', 'gene_B', 'target']
        value_cols = 'LRScore'
        data = data.loc[data['LRScore'].abs().nlargest(min(len(data), threshold)).index]
        title = f'{target} ligand gene interactions EXP vs CTR'

        gen_sankey(data, cat_cols, value_cols, title)
    
    else:
        print(f"Gene->{target} Not Found")
    

def gen_sankey2(df, cat_cols=[], value_cols='', title='Sankey Diagram'):
    """
    Helper function to the function plot_sankey()

     Parameters
    ----------
    df :
        Dataframe

    cat_cols :
        Columns interested in the sankey plot

    value_cols :
        Sankey plot generated using connections based on this value_cols

    title :
        Title of Sankey plot
    
    Returns
    -------
    Nothing (plots Sankey plot)

    """

    df['source'] += 'S'
    df['target'] += 'T'
    
    labelList = []
    for catCol in cat_cols:
        labelListTemp =  list((df[catCol].values))
        labelList = labelList + labelListTemp    
        
    for i in range(len(cat_cols)-1):
        if i==0:
            sourceTargetDf = df[[cat_cols[i],cat_cols[i+1],value_cols]]
            sourceTargetDf.columns = ['source','target','count']
        else:
            tempDf = df[[cat_cols[i],cat_cols[i+1],value_cols]]
            tempDf.columns = ['source','target','count']
            sourceTargetDf = pd.concat([sourceTargetDf,tempDf])
        # sourceTargetDf = sourceTargetDf.groupby(['source','target']).agg({'count':'sum'}).reset_index()
        
    sourceTargetDf['sourceID'] = sourceTargetDf['source'].apply(lambda x: labelList.index(x))
    sourceTargetDf['targetID'] = sourceTargetDf['target'].apply(lambda x: labelList.index(x))

    for i, label in enumerate(labelList):
        if label[-1:] == 'S' or label[-1:] == 'T':
            labelList[i] = labelList[i][:-1]
        
    norm = mcolors.Normalize(vmin=min(sourceTargetDf['count']), vmax=max(sourceTargetDf['count']))
    colormap = cm.get_cmap('RdBu_r')
    link_colors = [mcolors.to_hex(colormap(norm(value))) for value in sourceTargetDf['count']]
    
    fig =  go.Figure(data = [go.Sankey(
        node = dict(
          pad = 0,
          thickness = 20,
          line = dict(
            color = "black",
            width = 0.5
          ),
          label = labelList,
          color = "white"
        ),
        link = dict(
          source = sourceTargetDf['sourceID'],
          target = sourceTargetDf['targetID'],
          value = [abs(i) for i in sourceTargetDf['count']],
          color = link_colors
        )    
    )])
    
    colorbar_trace = go.Scatter(
        x=[None], y=[None], mode='markers',
        marker=dict(
            colorscale='RdBu_r',
            cmin=min(sourceTargetDf['count']),
            cmax=max(sourceTargetDf['count']),
            colorbar=dict(
                title="Value",
                thickness=15,
                len=0.5,
                x=1.05,
                xref="paper"
            )
        ),
        hoverinfo='none'
    )

    fig.add_trace(colorbar_trace)

    fig.add_annotation(x=0, y=1.05, yref="paper", text="Source", showarrow=False, font=dict(size=10))
    fig.add_annotation(x=0.33, y=1.05, yref="paper", text="Ligand", showarrow=False, font=dict(size=10))
    fig.add_annotation(x=0.66, y=1.05, yref="paper", text="Receptor", showarrow=False, font=dict(size=10))
    fig.add_annotation(x=1, y=1.05, yref="paper", text="Target", showarrow=False, font=dict(size=10))

    fig.update_layout(
        xaxis=dict(showgrid=False, zeroline=False, showticklabels=False, range=[0,1]),
        yaxis=dict(showgrid=False, zeroline=False, showticklabels=False, range=[0,1]),
        plot_bgcolor='white',
        autosize=True,
        width = None,
        height = 600,
        title = title,
        font = dict(size=10)
        )
    
    fig.show(config={"responsive": True})

def gen_sankey(df, cat_cols=[], value_cols='', title='Sankey Diagram'):
    """
    Helper function to the function plot_sankey()

     Parameters
    ----------
    df :
        Dataframe

    cat_cols :
        Columns interested in the sankey plot

    value_cols :
        Sankey plot generated using connections based on this value_cols

    title :
        Title of Sankey plot
    
    Returns
    -------
    Nothing (plots Sankey plot)

    """

    # df['source'] += 'S'
    df['target'] += ' '
    
    labelList = []
    for catCol in cat_cols:
        labelListTemp =  list((df[catCol].values))
        labelList = labelList + labelListTemp    
        
    for i in range(len(cat_cols)-1):
        if i==0:
            sourceTargetDf = df[[cat_cols[i],cat_cols[i+1],value_cols]]
            sourceTargetDf.columns = ['source','target','count']
        else:
            tempDf = df[[cat_cols[i],cat_cols[i+1],value_cols]]
            tempDf.columns = ['source','target','count']
            sourceTargetDf = pd.concat([sourceTargetDf,tempDf])
    
    vmin = sourceTargetDf['count'].min()
    vmax = sourceTargetDf['count'].max()
    limit = max(abs(vmin), abs(vmax))
    vcenter = 0
    norm = mcolors.TwoSlopeNorm(vmin=-limit, vcenter=vcenter, vmax=limit)
    # norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
    cmap = plt.get_cmap('RdBu_r')
    sourceTargetDf['hex_color'] = sourceTargetDf['count'].apply(lambda x: mcolors.to_hex(cmap(norm(x))))
    
    flows = []
    for i, row in sourceTargetDf.iterrows():
        flows.append((row['source'], row['target'], 1, {'color': row['hex_color']}))

    nodes = Sankey.infer_nodes(flows)
    nodes_new = []
    for level in nodes:
        level_new = []
        for node in level:
            node_new = node + [{'color' : 'black',
                                'label_pos':'center', 'label_opts': dict(fontsize=10, bbox=dict(boxstyle='round,pad=0.3', edgecolor='black', facecolor='white'))}]
            level_new.append(node_new)
        nodes_new.append(level_new)

    fig, ax = plt.subplots(figsize=(15, 10))
    s = Sankey(flows=flows,
               nodes=nodes_new,
               flow_color_mode_alpha=0.3,
               node_opts=dict(label_format='{label}'),
    )
    s.draw(ax=ax)

    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax, orientation='vertical', pad=0.01, shrink=0.5)
    cbar.set_label(value_cols, fontsize=12)

    ax.text(x=-0.05, y=1.02, s="Source", fontsize=10)
    ax.text(x=0.95, y=1.02, s="Ligand", fontsize=10)
    ax.text(x=1.95, y=1.02, s="Receptor", fontsize=10)
    ax.text(x=2.95, y=1.02, s="Target", fontsize=10)

    ax.set_title(title)
    plt.tight_layout()
    plt.show()

def gene_annotation(gene_list_to_profile,
                    num_gos: int = 15,
                    figsize=(10,6),
                    title: str = None,
                    font_size: int = 10,
                    organism: str = 'hsapiens',
                    dpi: int = 100,
                    s: int = 100,
                    color: str = 'tab:blue'):
    """
    Perform Gene Ontology (GO) enrichment analysis and create a scatterplot of enriched terms.

    Parameters:
    ----------
    num_gos: int, optional
        Number of GO terms to plot. Default is 5.
    figsize: tuple, optional
        figsize. Default is (6,6).
    title: str
        Title of the plot.
    font_size: int, optional
        Font size for labels. Default is 10.
    0rganism: str, optional
        The organism for GO analysis. Default is 'hsapiens'.
    dpi: int, optional
        Dots per inch for the saved plot image. Default is 100.
    s: int, optional
        Marker size for the scatterplot. Default is 200.
    color: str, optional
        Color of the scatterplot markers. Default is 'tab:blue'.

    Returns:
    --------
    None
        Plots the scatterplot of enriched GO terms.
    """

    
    gp = GProfiler(return_dataframe=True)
    if gene_list_to_profile:
        gprofiler_results = gp.profile(organism = organism,
                                       query = gene_list_to_profile)
    else:
        return "Genes list is empty!"
    
    
    if(gprofiler_results.shape[0] == 0):
        return "Not enough information!"

    
    if(gprofiler_results.shape[0] < num_gos):
        num_gos = gprofiler_results.shape[0]

  
    selected_gps = gprofiler_results.head(num_gos)[['name', 'p_value']]
    
    selected_gps['nlog10'] = -np.log10(selected_gps['p_value'].values)

    plt.figure(figsize = figsize, dpi = dpi)
    # plt.style.use('default')
    sns.scatterplot(data = selected_gps, x = "nlog10", y = "name", s = s, color = color)

    plt.title(title, fontsize = font_size)

    plt.xticks(size = font_size)
    plt.yticks(size = font_size)

    plt.ylabel("GO Terms", size = font_size)
    plt.xlabel("-$log_{10}$ (P-value)", size = font_size)

    plt.tight_layout()
    plt.show()


def plot_volcane(df):
    """
    This function generates a Volcano plot

    Parameters
    ----------
    df :
        Dataframe

    Returns
    -------
    Python default volcano plot

    """
    np.random.seed(42)
    data = df
    data['neg_log10_p_value'] = -np.log10(df['p_value'])
    
    p_threshold = 0.05
    fc_threshold = 1

    data["color"] = "gray"
    data.loc[(data["lodds"] > fc_threshold) & (data["p_value"] < p_threshold), "color"] = "red"
    data.loc[(data["lodds"] < -fc_threshold) & (data["p_value"] < p_threshold), "color"] = "red"
    data.loc[(data["lodds"] > -fc_threshold) & (data["lodds"] < fc_threshold) & (data["p_value"] < p_threshold), "color"] = "blue"
    data.loc[(data["lodds"] < -fc_threshold) & (data["p_value"] > p_threshold), "color"] = "green"
    data.loc[(data["lodds"] > fc_threshold) & (data["p_value"] > p_threshold), "color"] = "green"

    # Plot
    plt.figure(figsize=(8, 6))
    sns.scatterplot(x="lodds", y="neg_log10_p_value", hue="color", palette={"gray": "gray", "red": "red", "blue": "blue", "green": "green"}, data=data, edgecolor=None, alpha=0.7)

    # Add significance threshold lines
    plt.axhline(-np.log10(p_threshold), linestyle="--", color="black", linewidth=1)  # P-value threshold
    plt.axvline(fc_threshold, linestyle="--", color="black", linewidth=1)  # Positive log2FC threshold
    plt.axvline(-fc_threshold, linestyle="--", color="black", linewidth=1)  # Negative log2FC threshold

    for i, row in data.iterrows():
        if row['color'] == 'red':
            plt.text(row["lodds"], row["neg_log10_p_value"], row["cellpair"], fontsize=8, ha='right')

    x_limit = max(abs(data["lodds"].min()), abs(data["lodds"].max()))
    plt.xlim(-x_limit-1, x_limit+1)

    plt.xlabel(r"Log$_{2}$ Fold Change")
    plt.ylabel(r"-Log$_{10}$(p-value)")
    plt.title("Volcano Plot")
    plt.legend([],[], frameon=False)  # Hide legend
    plt.show()


def plot_clustermap(data, title):
    # Create pivot table for Gene_A-Gene_B without aggregation
    pivot_table = data.groupby(["source", "target"])["LRScore"].sum().unstack().fillna(0)
    xlabel, ylabel = "Target Tissue", "Source Tissue"

    # Plot heatmap
    g = sns.clustermap(
        pivot_table,
        figsize=(9, 7),
        annot=True,
        linewidths=0.5,
        method="ward",    # Clustering method (options: single, complete, average, ward)
        metric="euclidean",  # Distance metric (options: euclidean, cityblock, cosine, etc.)
        dendrogram_ratio=(0.2, 0.2),  # Adjust dendrogram size
        cbar_pos=(0.02, 0.8, 0.03, 0.15)  # Adjust colorbar position
    )

    # Set Labels
    g.ax_heatmap.set_xlabel(xlabel)
    g.ax_heatmap.set_ylabel(ylabel)
    plt.title(title, fontsize=14)

    plt.show()




__version__ = "1.0.1"
__author__ = 'Ruthvik Koppala'
__credits__ = 'Institute for Computational Genomics'

import pycrosstalker.tools as tl
import pycrosstalker.plots as pl


#Atfan helped to translate some of the code